

#include <AudioUnit/AudioUnit.h>
#include <AudioToolbox/AudioToolbox.h>
#include <CoreFoundation/CFURL.h>
#include <OpenAL/al.h>
#include <OpenAL/alc.h>
#include <pthread.h>
#include <mach/mach.h>

#include "StreamSound.h"

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
typedef ALvoid AL_APIENTRY ( * alBufferDataStaticProcPtr) ( const ALint bid, ALenum format, ALvoid* data, ALsizei size, ALsizei freq);
static ALvoid alBufferDataStaticProc( const ALint bid, ALenum format, ALvoid* data, ALsizei size, ALsizei freq)
{
    static alBufferDataStaticProcPtr proc = NULL;
    
    if ( proc == NULL) 
    {
        proc = ( alBufferDataStaticProcPtr) alcGetProcAddress( NULL, ( const ALCchar *) "alBufferDataStatic");
    }
    
    if ( proc)
    {    
        proc( bid, format, data, size, freq);
    }

    return;
}


OggStream::OggStream()
  : mOggFile( NULL),
    mVorbisInfo( NULL),
    mVorbisComment( NULL),
    mBufferIndex( 0),
    mSource( 0),
    mFormat( 0)
{
    int oIndex = 0;
    while ( oIndex < kBufferCount)
    {
        mBuffers[ oIndex] = 0;
        oIndex ++;
    }
}

OggStream::~OggStream()
{
}

void OggStream::open( const char *inPath)
{
    int oResult;
    
    mOggFile = fopen( inPath, "rb");
    if ( mOggFile == NULL)
    {
        printf( "Could not open Ogg file.");
        return;
    }

    oResult = ov_open( mOggFile, &mOggStream, NULL, 0);
    if ( oResult < 0)
    {
        fclose( mOggFile);
        printf( "Could not open Ogg stream. %s\n", errorString( oResult));
        return;
    }

    mVorbisInfo = ov_info( &mOggStream, -1);
    mVorbisComment = ov_comment( &mOggStream, -1);

    if ( mVorbisInfo->channels == 1)
    {
        mFormat = AL_FORMAT_MONO16;
    }
    else
    {
        mFormat = AL_FORMAT_STEREO16;
    }
        
    alGenBuffers( kBufferCount, mBuffers);
    checkError();
    alGenSources( 1, &mSource);
    checkError();

    alSource3f( mSource, AL_POSITION, 0.0, 0.0, 0.0);
    alSource3f( mSource, AL_VELOCITY, 0.0, 0.0, 0.0);
    alSource3f( mSource, AL_DIRECTION, 0.0, 0.0, 0.0);
    alSourcef( mSource, AL_ROLLOFF_FACTOR, 0.0);
    alSourcei( mSource, AL_SOURCE_RELATIVE, AL_TRUE);

    mPlayThread = new StreamThread( playbackProc, ( void *) this);
    mPlayThread->Start();
}

void OggStream::release()
{
    printf( "RELEASE\n");
    alSourceStop( mSource);
    clearBuffers();
    alDeleteSources( 1, &mSource);
    checkError();
    alDeleteBuffers( kBufferCount, mBuffers);
    checkError();

    ov_clear( &mOggStream);
}

void OggStream::display()
{
    printf( "version         %d\n", mVorbisInfo->version);
    printf( "channels        %d\n", mVorbisInfo->channels);
    printf( "rate (hz)       %d\n", ( int) mVorbisInfo->rate);
    printf( "bitrate upper   %d\n", ( int) mVorbisInfo->bitrate_upper);
    printf( "bitrate nominal %d\n", ( int) mVorbisInfo->bitrate_nominal);
    printf( "bitrate lower   %d\n", ( int) mVorbisInfo->bitrate_lower);
    printf( "bitrate window  %d\n", ( int) mVorbisInfo->bitrate_window);
    printf( "\n");
    printf( "vendor %s\n", mVorbisComment->vendor);
        
    for(int i = 0; i < mVorbisComment->comments; i++)
        printf( "   %s\n", mVorbisComment->user_comments[i]);
    printf( "\n");
}

void *OggStream::playbackProc( void *inData)
{
    OggStream *THIS = ( OggStream *) inData;

    if ( THIS->playback() == false)
    {
        printf( "Ogg refused to play.\n");
        return NULL;
    }

    while ( THIS->update() == true)
    {
        if ( THIS->playing() == false)
        {
            if ( THIS->playback() == false)
            {
                printf( "Ogg abruptly stopped.\n");
                return NULL;
            }
            else
            {
                printf( "Ogg stream was interrupted.\n");
            }
        }
        sleep( 1000);
    }

    printf( "DONE WITH LOOP.\n");
    THIS->release();
    return NULL;
}

bool OggStream::playback()
{
    if ( playing() == true)
    {
        return true;
    }

    int oIndex = 0;
    while ( oIndex < kBufferCount)
    {
        printf( "INIT BUFFER %x\n", mBuffers[ oIndex]);
        if ( stream( mBuffers[ oIndex]) == 0)
        {
            return false;
        }
        oIndex ++;
    }

    alSourceQueueBuffers( mSource, kBufferCount, mBuffers);
    alSourcePlay( mSource);

    return true;
}

bool OggStream::playing()
{
    ALenum oState;
    alGetSourcei( mSource, AL_SOURCE_STATE, &oState);
    return ( oState == AL_PLAYING);
}

bool OggStream::update()
{
    int oProcessed;
    bool obActive = true;

    alGetSourcei( mSource, AL_BUFFERS_PROCESSED, &oProcessed);
//  printf( "UPDATE %d %d\n", mSource, oProcessed);

    while( oProcessed--)
    {
        ALuint oBuffer;
        
        alSourceUnqueueBuffers( mSource, 1, &oBuffer);
        checkError();

        printf( "STREAM TO BUFFER %x\n", oBuffer);
        obActive = stream( oBuffer);

        alSourceQueueBuffers( mSource, 1, &oBuffer);
        checkError();
    }

    return obActive;
}


bool OggStream::stream( ALuint inBuffer)
{
    int oSize = 0;
    int oSection;
    int oResult;
    char *oPCM = mPCMBuffers[ mBufferIndex];
//  printf( "STREAM\n");
    while( oSize < BUFFER_SIZE)
    {
        oResult = ov_read( &mOggStream, oPCM + oSize, BUFFER_SIZE - oSize, 0, 2, 1, &oSection);
//      printf( "   READ %d %d %d\n", oResult, BUFFER_SIZE, oSection);
    
        if ( oResult > 0)
        {
            oSize += oResult;
        }
        else
        {
            if ( oResult < 0)
            {    
                printf( "%s\n", errorString( oResult));
                return false;
            }
            else
            {    
                break;
            }
        }
    }
    
    if ( oSize == 0)
    {
        return false;
    }

    printf( "END STREAM 1 %x %x %x %d %d\n", inBuffer, mFormat, ( unsigned int) oPCM, oSize, ( int) mVorbisInfo->rate);
//  alBufferDataStaticProc( inBuffer, mFormat, oPCM, oSize, mVorbisInfo->rate);
    alBufferData( inBuffer, mFormat, oPCM, oSize, mVorbisInfo->rate);
    checkError();
    printf( "END STREAM 2 %d %d\n", oSize, BUFFER_SIZE);
    
    mBufferIndex ++;
    if ( mBufferIndex == kBufferCount)
    {
        mBufferIndex = 0;
    }

    return true;
}

void OggStream::clearBuffers()
{
    int oQueued;
    
    alGetSourcei( mSource, AL_BUFFERS_QUEUED, &oQueued);
    while ( oQueued--)
    {
        ALuint oBuffer;
        alSourceUnqueueBuffers( mSource, 1, &oBuffer);
        checkError();
    }
}

void OggStream::checkError()
{
    int oError = alGetError();
    if ( oError != AL_NO_ERROR)
    {
        const char *oString = "Unknown.";
        switch( oError)
        {
            case AL_NO_ERROR:
                oString = "There is no current error.";
                break;

            case AL_INVALID_NAME:
                oString = "Invalid name parameter.";
                break;

            case AL_INVALID_ENUM:
                oString = "Invalid parameter.";
                break;

            case AL_INVALID_VALUE:
                oString = "Invalid enum parameter value.";
                break;

            case AL_INVALID_OPERATION:
                oString = "Illegal call.";
                break;

            case AL_OUT_OF_MEMORY:
                oString = "Unable to allocate memory.";
                break;
        }
         

        printf( "OpenAL error was raised. %d %s\n", oError, oString);
    }
}

const char *OggStream::errorString( int inCode)
{
    switch( inCode)
    {
        case OV_EREAD:
            return "Read from media.";
        case OV_ENOTVORBIS:
            return "Not Vorbis data.";
        case OV_EVERSION:
            return "Vorbis version mismatch.";
        case OV_EBADHEADER:
            return "Invalid Vorbis header.";
        case OV_EFAULT:
            return "Internal logic fault (bug or heap/stack corruption.";
        default:
            return "Unknown Ogg error.";
    }
    return "Unknown Ogg error.";
}



StreamThread::StreamThread( ThreadRoutine inThreadRoutine, void* inParameter)
    :   mPThread( 0),
        mSpawningThreadPriority( getScheduledPriority( pthread_self(), StreamThread_SET_PRIORITY)),
        mThreadRoutine( inThreadRoutine),
        mThreadParameter( inParameter),
        mPriority( kDefaultThreadPriority),
        mFixedPriority( false),
        mAutoDelete( true) 
{
}

StreamThread::~StreamThread() 
{
}

bool StreamThread::IsRunning() const 
{ 
    return 0 != mPThread; 
}

void StreamThread::SetAutoDelete( bool inbAutoDelete) 
{ 
    mAutoDelete = inbAutoDelete; 
}

void StreamThread::SetPriority( UInt32 inPriority, bool inFixedPriority)
{
    OSStatus result = noErr;
    mPriority = inPriority;
    mFixedPriority = inFixedPriority;
    if( mPThread != 0)
    {
        if ( mFixedPriority)
        {
            thread_extended_policy_data_t theFixedPolicy;
            theFixedPolicy.timeshare = false;   // set to true for a non-fixed thread

            result  = thread_policy_set( pthread_mach_thread_np( mPThread), THREAD_EXTENDED_POLICY, ( thread_policy_t) &theFixedPolicy, THREAD_EXTENDED_POLICY_COUNT);
            if ( result) 
            {
                printf( "StreamThread::SetPriority: failed to set the fixed-priority policy");
                return;
            }
        }

        // We keep a reference to the spawning thread's priority around (initialized in the constructor), 
        // and set the importance of the child thread relative to the spawning thread's priority.
        thread_precedence_policy_data_t thePrecedencePolicy;
        
        thePrecedencePolicy.importance = mPriority - mSpawningThreadPriority;
        result = thread_policy_set( pthread_mach_thread_np( mPThread), THREAD_PRECEDENCE_POLICY, ( thread_policy_t) &thePrecedencePolicy, THREAD_PRECEDENCE_POLICY_COUNT);
        if ( result) 
        {
            printf( "StreamThread::SetPriority: failed to set the precedence policy");
            return;
        }
    } 
}

//  Actions
void StreamThread::Start()
{
    if ( mPThread != 0)
    {
        printf( "StreamThread::Start: can't start because the thread is already running\n");
        return;
    }

    pthread_attr_t theThreadAttributes;
    
    pthread_attr_init( &theThreadAttributes);
    pthread_attr_setdetachstate( &theThreadAttributes, PTHREAD_CREATE_DETACHED);
    pthread_create( &mPThread, &theThreadAttributes, ( ThreadRoutine) StreamThread::Entry, this);
    pthread_attr_destroy( &theThreadAttributes);
}

void *StreamThread::Entry( StreamThread* inStreamThread)
{
    void *theAnswer = NULL;

    inStreamThread->SetPriority( inStreamThread->mPriority, inStreamThread->mFixedPriority);

    if ( inStreamThread->mThreadRoutine != NULL)
    {
        theAnswer = inStreamThread->mThreadRoutine( inStreamThread->mThreadParameter);
    }

    inStreamThread->mPThread = 0;
    if ( inStreamThread->mAutoDelete)
    {
        delete inStreamThread;
    }
    return theAnswer;
}

UInt32 StreamThread::getScheduledPriority( pthread_t inThread, int inPriorityKind)
{
    thread_basic_info_data_t threadInfo;
    policy_info_data_t thePolicyInfo;
    unsigned int count;

    if ( inThread == NULL)
    {
        return 0;
    }
    
    // get basic info
    count = THREAD_BASIC_INFO_COUNT;
    thread_info( pthread_mach_thread_np( inThread), THREAD_BASIC_INFO, ( thread_info_t) &threadInfo, &count);
    
    switch ( threadInfo.policy) 
    {
        case POLICY_TIMESHARE:
            count = POLICY_TIMESHARE_INFO_COUNT;
            thread_info( pthread_mach_thread_np ( inThread), THREAD_SCHED_TIMESHARE_INFO, ( thread_info_t) &( thePolicyInfo.ts), &count);
            if ( inPriorityKind == StreamThread_SCHEDULED_PRIORITY) 
            {
                return thePolicyInfo.ts.cur_priority;
            }
            return thePolicyInfo.ts.base_priority;
            break;
            
        case POLICY_FIFO:
            count = POLICY_FIFO_INFO_COUNT;
            thread_info( pthread_mach_thread_np( inThread), THREAD_SCHED_FIFO_INFO, ( thread_info_t) &( thePolicyInfo.fifo), &count);
            if ( (thePolicyInfo.fifo.depressed) && (inPriorityKind == StreamThread_SCHEDULED_PRIORITY) ) 
            {
                return thePolicyInfo.fifo.depress_priority;
            }
            return thePolicyInfo.fifo.base_priority;
            break;
            
        case POLICY_RR:
            count = POLICY_RR_INFO_COUNT;
            thread_info( pthread_mach_thread_np( inThread), THREAD_SCHED_RR_INFO, ( thread_info_t) &( thePolicyInfo.rr), &count);
            if (( thePolicyInfo.rr.depressed) && ( inPriorityKind == StreamThread_SCHEDULED_PRIORITY))
            {
                return thePolicyInfo.rr.depress_priority;
            }
            return thePolicyInfo.rr.base_priority;
            break;
    }

    return 0;
}



AudioQueueOggStream::AudioQueueOggStream() 
    :   mQueue( 0),
        mBufferByteSize( BUFFER_SIZE),
        mVolume( 1.0)
{   
    mFileFormat.mSampleRate = 44100;
    mFileFormat.mFormatID = kAudioFormatLinearPCM;
    mFileFormat.mFormatFlags = kAudioFormatFlagIsSignedInteger | kLinearPCMFormatFlagIsBigEndian | kAudioFormatFlagIsPacked; 
    mFileFormat.mChannelsPerFrame = 1;
    mFileFormat.mBitsPerChannel = 16;
    mFileFormat.mBytesPerPacket = mFileFormat.mBytesPerFrame = ( mFileFormat.mChannelsPerFrame * sizeof( SInt16)); 
    mFileFormat.mFramesPerPacket = 1;
    mFileFormat.mReserved = 0;

}

AudioQueueOggStream::~AudioQueueOggStream() 
{ 
    Teardown(); 
}

void AudioQueueOggStream::Teardown()
{
    ov_clear( &mOggStream);
    if ( mQueue != NULL)
    {
        AudioQueueDispose( mQueue, true);
    }
}

const char *AudioQueueOggStream::errorString( int inCode)
{
    switch( inCode)
    {
        case OV_EREAD:
            return "Read from media.";
        case OV_ENOTVORBIS:
            return "Not Vorbis data.";
        case OV_EVERSION:
            return "Vorbis version mismatch.";
        case OV_EBADHEADER:
            return "Invalid Vorbis header.";
        case OV_EFAULT:
            return "Internal logic fault (bug or heap/stack corruption.";
        default:
            return "Unknown Ogg error.";
    }
    return "Unknown Ogg error.";
}

void AudioQueueOggStream::display()
{
    printf( "version         %d\n", mVorbisInfo->version);
    printf( "channels        %d\n", mVorbisInfo->channels);
    printf( "rate (hz)       %d\n", ( int) mVorbisInfo->rate);
    printf( "bitrate upper   %d\n", ( int) mVorbisInfo->bitrate_upper);
    printf( "bitrate nominal %d\n", ( int) mVorbisInfo->bitrate_nominal);
    printf( "bitrate lower   %d\n", ( int) mVorbisInfo->bitrate_lower);
    printf( "bitrate window  %d\n", ( int) mVorbisInfo->bitrate_window);
    printf( "\n");
    printf( "vendor %s\n", mVorbisComment->vendor);
        
    for(int i = 0; i < mVorbisComment->comments; i++)
        printf( "   %s\n", mVorbisComment->user_comments[i]);
    printf( "\n");
}

void AudioQueueOggStream::open( const char *inPath)
{
    int oResult;
    
    mOggFile = fopen( inPath, "rb");
    if ( mOggFile == NULL)
    {
        printf( "Could not open Ogg file.");
        return;
    }

    oResult = ov_open( mOggFile, &mOggStream, NULL, 0);
    if ( oResult < 0)
    {
        fclose( mOggFile);
        printf( "Could not open Ogg stream. %s\n", errorString( oResult));
        return;
    }

    mVorbisInfo = ov_info( &mOggStream, -1);
    mVorbisComment = ov_comment( &mOggStream, -1);

    if ( mVorbisInfo->channels == 1)
    {
        mFormat = AL_FORMAT_MONO16;
    }
    else
    {
        mFormat = AL_FORMAT_STEREO16;
    }
        
    // init code
    SetupQueue();
    SetupBuffers();
}

void AudioQueueOggStream::QueueStoppedProc( void *inUserData, AudioQueueRef inAQ, AudioQueuePropertyID inID)
{
    printf( "QUEUE STOPPED\n");
//  UInt32 isRunning;
//  UInt32 propSize = sizeof( isRunning);
//  BackgroundTrackMgr *THIS = ( BackgroundTrackMgr *) inUserData;
//  AudioQueueGetProperty( inAQ, kAudioQueueProperty_IsRunning, &isRunning, &propSize);
}

Boolean AudioQueueOggStream::DisposeBuffer( AudioQueueRef inAQ, std::vector<AudioQueueBufferRef> inDisposeBufferList, AudioQueueBufferRef inBufferToDispose)
{
    for ( unsigned int i = 0; i < inDisposeBufferList.size(); i++)
    {
        if ( inBufferToDispose == inDisposeBufferList[ i])
        {
            AudioQueueFreeBuffer( inAQ, inBufferToDispose);
            inDisposeBufferList.pop_back();
            return true;
        }
    }
    return false;
}

void AudioQueueOggStream::QueueCallback( void *inUserData, AudioQueueRef inAQ, AudioQueueBufferRef inCompleteAQBuffer)
{
    // dispose of the buffer if no longer in use
    AudioQueueOggStream *THIS = ( AudioQueueOggStream *) inUserData;
    if ( DisposeBuffer( inAQ, THIS->mBuffersToDispose, inCompleteAQBuffer))
    {    
        return;
    }
    
    int oSize = 0;
    int oSection;
    int oResult;
    char *oBuffer = ( char *) inCompleteAQBuffer->mAudioData;
    while( oSize < BUFFER_SIZE)
    {
        oResult = ov_read( &THIS->mOggStream, oBuffer + oSize, BUFFER_SIZE - oSize, 0, 2, 1, &oSection);
        if ( oResult > 0)
        {
            oSize += oResult;
        }
        else
        {
            if ( oResult < 0)
            {    
                printf( "READ OGG NEG RESULT\n");
                AudioQueueStop( inAQ, false);
                return;
            }
            else
            {    
                break;
            }
        }
    }
    
    if ( oSize == 0)
    {
        printf( "FILE DONE\n");
        AudioQueueStop( inAQ, false);
        return ;
    }

    inCompleteAQBuffer->mAudioDataByteSize = oSize;  
    AudioQueueEnqueueBuffer( inAQ, inCompleteAQBuffer, 0, 0);
}

void AudioQueueOggStream::SetupQueue()
{
//  AudioQueueNewOutput( &mFileFormat, QueueCallback, this, CFRunLoopGetCurrent(), kCFRunLoopCommonModes, 0, &mQueue);
    AudioQueueNewOutput( &mFileFormat, QueueCallback, this, NULL, NULL, 0, &mQueue);

    // add a notification proc for when the queue stops
    AudioQueueAddPropertyListener( mQueue, kAudioQueueProperty_IsRunning, QueueStoppedProc, this);
    SetVolume( mVolume);
}

void AudioQueueOggStream::SetupBuffers()
{
    // allocate the queue's buffers
    for ( int i = 0; i < kNumberBuffers ; ++i) 
    {
        AudioQueueAllocateBuffer( mQueue, mBufferByteSize, &mBuffers[ i]);
        QueueCallback( this, mQueue, mBuffers[ i]);
    }
}

void AudioQueueOggStream::Start()
{
    AudioQueuePrime( mQueue, 1, NULL); 
    AudioQueueStart( mQueue, NULL);
}

void AudioQueueOggStream::Stop()
{
    AudioQueueStop( mQueue, true);
}

void AudioQueueOggStream::SetVolume( Float32 inVolume)
{
    mVolume = inVolume;
    AudioQueueSetParameter( mQueue, kAudioQueueParam_Volume, mVolume);
}







#define kOutputBus  0
#define kInputBus   1

AudioUnitOggStream *AudioUnitOggStream::sInstance = NULL;

void AudioUnitOggStream::streamOgg( const char *inFileName)
{
    if ( sInstance == NULL)
    {
        sInstance = new AudioUnitOggStream();
    }
    memcpy( sInstance->mFileName, inFileName, strlen( inFileName) + 1);
    sInstance->mbLoadFile = TRUE;
}

void AudioUnitOggStream::stopOgg()
{
    if ( sInstance != NULL)
    {
        sInstance->mbCloseFile = TRUE;
    }
}

AudioUnitOggStream::AudioUnitOggStream() 
  : mReadBufferIndex( 0),
    mWriteBufferIndex( 0),
    mOggFile( NULL),
    mbLoadFile( FALSE),
    mbCloseFile( FALSE),
    mbAudioUnitRunning( FALSE),
    mbThreadRunning( FALSE),
    mVorbisInfo( NULL),
    mVorbisComment( NULL)
{   
    int oIndex = 0;
    while ( oIndex < STREAM_BUFFER_COUNT)
    {
        mBufferArray[ oIndex].mBufferState = kOggStreamBufferEmpty;
        mBufferArray[ oIndex].mBufferSize = 0;
        mBufferArray[ oIndex].mReadPosition = 0;
        oIndex ++;
    }

    // Describe audio component
    AudioComponentDescription oAudioComponentDescription;
    oAudioComponentDescription.componentType = kAudioUnitType_Output;
    oAudioComponentDescription.componentSubType = kAudioUnitSubType_RemoteIO;
    oAudioComponentDescription.componentFlags = 0;
    oAudioComponentDescription.componentFlagsMask = 0;
    oAudioComponentDescription.componentManufacturer = kAudioUnitManufacturer_Apple;

    // Get component
    AudioComponent oInputComponent = AudioComponentFindNext( NULL, &oAudioComponentDescription);

    // Get audio units
    mStatus = AudioComponentInstanceNew( oInputComponent, &mAudioUnit);
    checkStatus();

    // Enable IO for recording
    UInt32 oFlag = 1;

    // Enable IO for playback
    mStatus = AudioUnitSetProperty( mAudioUnit, kAudioOutputUnitProperty_EnableIO, kAudioUnitScope_Output, kOutputBus, &oFlag, sizeof( oFlag));
    checkStatus();

    // Describe format
    mFileFormat.mSampleRate = 44100.00;
    mFileFormat.mFormatID = kAudioFormatLinearPCM;
    mFileFormat.mFormatFlags = kAudioFormatFlagIsSignedInteger | kAudioFormatFlagIsPacked;
    mFileFormat.mFramesPerPacket = 1;
    mFileFormat.mChannelsPerFrame = 1;
    mFileFormat.mBitsPerChannel = 16;
    mFileFormat.mBytesPerPacket = 2;
    mFileFormat.mBytesPerFrame = 2; 
    mFileFormat.mReserved = 0;

    // Apply format
    mStatus = AudioUnitSetProperty( mAudioUnit, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Input, kOutputBus, &mFileFormat, sizeof( mFileFormat));
    checkStatus();

    // Set output callback
    AURenderCallbackStruct oCallbackStruct;
    oCallbackStruct.inputProc = playbackCallback;
    oCallbackStruct.inputProcRefCon = this;
    mStatus = AudioUnitSetProperty( mAudioUnit, kAudioUnitProperty_SetRenderCallback, kAudioUnitScope_Global, kOutputBus, &oCallbackStruct, sizeof( oCallbackStruct));
    checkStatus();

    // Initialise
    mStatus = AudioUnitInitialize( mAudioUnit);
    checkStatus();

    mbThreadRunning = TRUE;
//  mPlayThread = new StreamThread( playbackProc, ( void *) this);
//  mPlayThread->Start();

    mBufferReadPos = 0;
    mBufferWritePos = 0;
    mBufferBytesWritten = 0;
    mBuffersWritten = 0;

    start();
}

AudioUnitOggStream::~AudioUnitOggStream() 
{ 
    teardown(); 
}

void AudioUnitOggStream::teardown()
{
    stop();
    close();
    mbThreadRunning = FALSE;
    AudioUnitUninitialize( mAudioUnit);
}


void AudioUnitOggStream::display()
{
    printf( "version         %d\n", mVorbisInfo->version);
    printf( "channels        %d\n", mVorbisInfo->channels);
    printf( "rate (hz)       %d\n", ( int) mVorbisInfo->rate);
    printf( "bitrate upper   %d\n", ( int) mVorbisInfo->bitrate_upper);
    printf( "bitrate nominal %d\n", ( int) mVorbisInfo->bitrate_nominal);
    printf( "bitrate lower   %d\n", ( int) mVorbisInfo->bitrate_lower);
    printf( "bitrate window  %d\n", ( int) mVorbisInfo->bitrate_window);
    printf( "\n");
    printf( "vendor %s\n", mVorbisComment->vendor);
        
    for ( int i = 0; i < mVorbisComment->comments; i++)
        printf( "   %s\n", mVorbisComment->user_comments[ i]);
    printf( "\n");
}

void *AudioUnitOggStream::playbackProc( void *inData)
{
    AudioUnitOggStream *opOggStream = ( AudioUnitOggStream *) inData;

    while ( opOggStream->writeDataToBuffer())
    {
        ;
    }

    return NULL;
}

void AudioUnitOggStream::run()
{
    if ( sInstance != NULL)
    {
//      sInstance->writeDataToBufferCircular();
        sInstance->writeDataToBuffer();
    }
}

bool AudioUnitOggStream::writeDataToBufferCircular()
{
    if ( mbLoadFile == TRUE)
    {
//      printf( "** OPEN:\n");
        mBufferBytesWritten = 0;
        mbLoadFile = FALSE;
        open( mFileName);
    }

    if (( mBufferWritePos >= STREAM_CIRCULAR_BUFFER_SIZE) &&
        ( mBufferReadPos != 0))
    {
//      printf( "** WRAP POS:\n");
        mBufferWritePos = 0;
    }

    // wait for the next buffer to be empty
    if (( mBufferWritePos == STREAM_CIRCULAR_BUFFER_SIZE) ||
        ( mBufferWritePos == ( mBufferReadPos - 1)) ||
        ( mOggFile == NULL))
    {
//      printf( "** NO WRITE:\n");
        return mbThreadRunning;
    }

    int oSize = mBufferReadPos - mBufferWritePos - 1;
    if ( mBufferReadPos <= mBufferWritePos)
    {
        oSize = STREAM_CIRCULAR_BUFFER_SIZE - mBufferWritePos;
    }
    if ( oSize != 0)
    {
        int oResult;
        int oSection;
//      printf( "** WRITE 1: %d %d %d\n", oSize, mBufferReadPos, mBufferWritePos);
        oResult = ov_read( &mOggStream, mCircularBuffer + mBufferWritePos, oSize, 0, 2, 1, &oSection);
        if ( oResult < 0)
        {
//          printf( "NEGATIVE OGG RESULT\n");
            oSize = 0;
        }
    
//      printf( "** WRITE 2: %d\n", oResult);

        if ( oResult == 0)
        {
//          printf( "** CLOSE FILE:\n");
            close();
        }
        mBufferWritePos += oResult;
        mBufferBytesWritten += oResult;
    }

    return mbThreadRunning;
}

bool AudioUnitOggStream::writeDataToBuffer()
{
    if ( mbLoadFile == TRUE)
    {
        mbLoadFile = FALSE;
        open( mFileName);
//      mBuffersWritten = 0;
//      printf( "** OPEN: %d %d\n", mWriteBufferIndex, mBufferArray[ mWriteBufferIndex].mBufferState);
    }

    // wait for the next buffer to be empty
    if (( mBufferArray[ mWriteBufferIndex].mBufferState != kOggStreamBufferEmpty) ||
        ( mOggFile == NULL))
    {
//      mPlayThread->SetPriority( StreamThread::kMinThreadPriority, false);
//      printf( "** NO WRITE: %d %d\n", mWriteBufferIndex, mBufferArray[ mWriteBufferIndex].mBufferState);
        return mbThreadRunning;
    }

//  mPlayThread->SetPriority( StreamThread::kMaxThreadPriority, false);
    mBufferArray[ mWriteBufferIndex].mBufferState = kOggStreamBufferWriting;

    int oSize = 0;
    int oSection;
    int oResult;
    while( oSize < STREAM_BUFFER_SIZE)
    {
        oResult = ov_read( &mOggStream, mBufferArray[ mWriteBufferIndex].mBuffer + oSize, STREAM_BUFFER_SIZE - oSize, 0, 2, 1, &oSection);
        if ( oResult > 0)
        {
            oSize += oResult;
        }
        else
        {
            if ( oResult < 0)
            {
                printf( "NEGATIVE OGG RESULT\n");
                oSize = 0;
            }
            break;
        }
//      sleep( 1000);
    }

    if ( oSize == 0)
    {
//      printf( "** CLOSE FILE: %d %d\n", mWriteBufferIndex, mBufferArray[ mWriteBufferIndex].mBufferState);
        mBufferArray[ mWriteBufferIndex].mBufferState = kOggStreamBufferEmpty;
        close();
    }
    else
    {
        mBuffersWritten ++;
        mBufferArray[ mWriteBufferIndex].mBufferSize = oSize;
        mBufferArray[ mWriteBufferIndex].mReadPosition = 0;
        mBufferArray[ mWriteBufferIndex].mBufferState = kOggStreamBufferFilled;
//      printf( "** WRITE: %d %d %d %x\n", mWriteBufferIndex, mBufferArray[ mWriteBufferIndex].mBufferState, oSize, * ( UInt32 *) ( mBufferArray[ mWriteBufferIndex].mBuffer));

        mWriteBufferIndex ++;
        if ( mWriteBufferIndex >= STREAM_BUFFER_COUNT)
        {
            mWriteBufferIndex = 0;
        }
    }
    return mbThreadRunning;
}

void AudioUnitOggStream::open( const char *inFileName)
{
    mOggFile = fopen( inFileName, "rb");
    if ( mOggFile == NULL)
    {
        printf( "Could not open Ogg file.");
        return;
    }

    int oResult = ov_open( mOggFile, &mOggStream, NULL, 0);
    if ( oResult < 0)
    {
        fclose( mOggFile);
        printf( "Could not open Ogg stream. %s\n", errorString( oResult));
        return;
    }

    mVorbisInfo = ov_info( &mOggStream, -1);
    mVorbisComment = ov_comment( &mOggStream, -1);

    if ( mVorbisInfo->channels == 1)
    {
        mFormat = AL_FORMAT_MONO16;
    }
    else
    {
        mFormat = AL_FORMAT_STEREO16;
    }
}

void AudioUnitOggStream::close()
{
    if ( mOggFile != NULL)
    {
        ov_clear( &mOggStream);
        mOggFile = NULL;
        mVorbisInfo = NULL;
        mVorbisComment = NULL;
    }
}

OSStatus AudioUnitOggStream::playbackCallback( void *inRefCon, AudioUnitRenderActionFlags *ioActionFlags, const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber, UInt32 inNumberFrames, AudioBufferList *ioData)
{
    AudioUnitOggStream *opOggStream = ( AudioUnitOggStream *) inRefCon;
//  opOggStream->readDataFromBufferCircular( inRefCon, ioActionFlags, inTimeStamp, inBusNumber, inNumberFrames, ioData);
    opOggStream->readDataFromBuffer( inRefCon, ioActionFlags, inTimeStamp, inBusNumber, inNumberFrames, ioData);
    return noErr;
}

void AudioUnitOggStream::readDataFromBufferCircular( void *inRefCon, AudioUnitRenderActionFlags *ioActionFlags, const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber, UInt32 inNumberFrames, AudioBufferList *ioData)
{
    // Notes: ioData contains buffers (may be more than one!)
    // Fill them up as much as you can. Remember to set the size value in each buffer to match how
    // much data is in the buffer.
    int oIndex = 0;
    while ( oIndex < ioData->mNumberBuffers)
    {
        // so we want to copy as much data as we can from the buffer to audio unit
        int oSize = mBufferWritePos - mBufferReadPos;
        if ( mBufferWritePos < mBufferReadPos)
        {
            oSize = STREAM_CIRCULAR_BUFFER_SIZE - mBufferReadPos;
        }

        if ( mBufferBytesWritten < STREAM_CIRCULAR_BUFFER_SIZE)
        {
            oSize = 0;
        }

//      printf( "** READ 1 %d %d %d:\n", oSize, mBufferWritePos, mBufferReadPos);
        if ( oSize > 0)
        {
            // is there enough room in the target buffer to fit oSize bytes?
            if ( ioData->mBuffers[ oIndex].mDataByteSize < oSize)
            {
                oSize = ioData->mBuffers[ oIndex].mDataByteSize;
            }
    
//          printf( "** READ 2 %d %d %d:\n", oSize, mBufferWritePos, mBufferReadPos);
            memcpy( ioData->mBuffers[ oIndex].mData, mCircularBuffer + mBufferReadPos, oSize);
            mBufferReadPos += oSize;
    
            if ( mBufferReadPos >= STREAM_CIRCULAR_BUFFER_SIZE)
            {
//              printf( "** WRAP READ:\n");
                mBufferReadPos = 0;
            }
        }

        ioData->mBuffers[ oIndex].mDataByteSize = oSize;
//      memset( (SInt16 *) ioData->mBuffers[ oIndex].mData, 0, ioData->mBuffers[ oIndex].mDataByteSize - oSize);
        oIndex ++;
    }
}

void AudioUnitOggStream::readDataFromBuffer( void *inRefCon, AudioUnitRenderActionFlags *ioActionFlags, const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber, UInt32 inNumberFrames, AudioBufferList *ioData)
{
    // wait for the next buffer to be empty
    if (( mBufferArray[ mReadBufferIndex].mBufferState != kOggStreamBufferFilled) ||
        ( mBuffersWritten < 2))
    {
//      printf( "READ 1: %d %d\n", mReadBufferIndex, mBufferArray[ mReadBufferIndex].mBufferState);
        // set the buffers to empty
        int oIndex = 0;
        while ( oIndex < ioData->mNumberBuffers)
        {
            memset(( SInt16 *) ioData->mBuffers[ oIndex].mData, 0, ioData->mBuffers[ oIndex].mDataByteSize);
//          ioData->mBuffers[ oIndex].mDataByteSize = 0;
            oIndex ++;
        }
        return ;
    }

//  printf( "READ 2 %x\n", * ( UInt32 *) ( mBufferArray[ mReadBufferIndex].mBuffer));
    // Notes: ioData contains buffers (may be more than one!)
    // Fill them up as much as you can. Remember to set the size value in each buffer to match how
    // much data is in the buffer.
    int oIndex = 0;
    while ( oIndex < ioData->mNumberBuffers)
    {
        // so we want to copy as much data as we can from the buffer to audio unit
        int oSize = mBufferArray[ mReadBufferIndex].mBufferSize - mBufferArray[ mReadBufferIndex].mReadPosition;
        if ( oSize > 0)
        {
            // is there enough room in the target buffer to fit oSize bytes?
            if ( ioData->mBuffers[ oIndex].mDataByteSize < oSize)
            {
                oSize = ioData->mBuffers[ oIndex].mDataByteSize;
            }
    
            memcpy( ioData->mBuffers[ oIndex].mData, mBufferArray[ mReadBufferIndex].mBuffer + mBufferArray[ mReadBufferIndex].mReadPosition, oSize);
            mBufferArray[ mReadBufferIndex].mReadPosition += oSize;
    
            // check if the buffer is fully read
            if ( mBufferArray[ mReadBufferIndex].mBufferSize == mBufferArray[ mReadBufferIndex].mReadPosition)
            {
                mBufferArray[ mReadBufferIndex].mReadPosition = 0;
                mBufferArray[ mReadBufferIndex].mBufferState = kOggStreamBufferEmpty;
//              printf( "READ 3: %d %d %x\n", mReadBufferIndex, oIndex, * ( UInt32 *) ( ioData->mBuffers[ oIndex].mData));
    
                mReadBufferIndex ++;
                if ( mReadBufferIndex >= STREAM_BUFFER_COUNT)
                {
                    mReadBufferIndex = 0;
                }
            }
        }

//      printf( "READ 3a: %d %d %d\n", oIndex, oSize, ioData->mBuffers[ oIndex].mDataByteSize);
//      ioData->mBuffers[ oIndex].mDataByteSize = oSize;
        memset(( SInt16 *) ioData->mBuffers[ oIndex].mData, 0, ioData->mBuffers[ oIndex].mDataByteSize - oSize);
        oIndex ++;
    }
    return ;
}

void AudioUnitOggStream::start()
{
    if ( mbAudioUnitRunning == FALSE)
    {
        // to start the system
        mbAudioUnitRunning = TRUE;
        mStatus = AudioOutputUnitStart( mAudioUnit);
        checkStatus();
    }
}

void AudioUnitOggStream::stop()
{
    if ( mbAudioUnitRunning == TRUE)
    {
        // to stop the system
        mbAudioUnitRunning = FALSE;
        mStatus = AudioOutputUnitStop( mAudioUnit);
        checkStatus();
    }
}

void AudioUnitOggStream::checkStatus()
{
    printf( "CHECK STATUS: %x\n", ( unsigned int) mStatus);
}

const char *AudioUnitOggStream::errorString( int inCode)
{
    switch( inCode)
    {
        case OV_EREAD:
            return "Read from media.";
        case OV_ENOTVORBIS:
            return "Not Vorbis data.";
        case OV_EVERSION:
            return "Vorbis version mismatch.";
        case OV_EBADHEADER:
            return "Invalid Vorbis header.";
        case OV_EFAULT:
            return "Internal logic fault (bug or heap/stack corruption.";
        default:
            return "Unknown Ogg error.";
    }
    return "Unknown Ogg error.";
}


