

#if !defined(__StreamSound_h__)
#define __StreamSound_h__

#include "vorbisfile.h"

#include <CoreAudio/CoreAudioTypes.h>
#include <AudioToolbox/AudioToolbox.h>
#include <OpenAL/al.h>
#include <OpenAL/alc.h>
#include <vector>

void streamSoundInit();
void streamSound( const char *inFileName);

class StreamThread;

#define BUFFER_SIZE         (4096 * 4)
#define kBufferCount        2

class OggStream
{
    public:

        OggStream();
        ~OggStream();

        void open( const char *inPath);
        void release();
        void display();
        bool playback();
        bool playing();
        bool update();

    protected:

        bool stream( ALuint inBuffer);
        void clearBuffers();
        void checkError();
        const char *errorString( int inCode);
        static void *playbackProc( void *inData);

    private:

        StreamThread *mPlayThread;

        FILE *mOggFile;
        OggVorbis_File mOggStream;
        vorbis_info *mVorbisInfo;
        vorbis_comment *mVorbisComment;

        int mBufferIndex;
        char mPCMBuffers[ kBufferCount][ BUFFER_SIZE];
        ALuint mBuffers[ kBufferCount];
        ALuint mSource;
        ALenum mFormat;
};

class StreamThread
{
    protected:

        typedef void *( * ThreadRoutine)( void *inParameter);

        // returns the thread's priority as it was last set by the API
        #define StreamThread_SET_PRIORITY               0

        // returns the thread's priority as it was last scheduled by the Kernel
        #define StreamThread_SCHEDULED_PRIORITY         1

        //  Member variables
        pthread_t mPThread;
        UInt32 mSpawningThreadPriority;
        ThreadRoutine mThreadRoutine;
        void *mThreadParameter;
        SInt32 mPriority;
        bool mFixedPriority;
        bool mAutoDelete;                           // delete self when thread terminates

    public:

        enum
        {
            kMinThreadPriority = 1,
            kMaxThreadPriority = 63,
            kDefaultThreadPriority = 31
        };

        StreamThread( ThreadRoutine inThreadRoutine, void* inParameter);
        ~StreamThread();
        bool IsRunning() const;
        void SetAutoDelete( bool inbAutoDelete);
        void SetPriority( UInt32 inPriority, bool inFixedPriority);
        void Start();

    protected:

        static void *Entry( StreamThread* inStreamThread);
        static UInt32 getScheduledPriority( pthread_t inThread, int inPriorityKind);

};

class AudioQueueOggStream
{   
    private:

        #define kNumberBuffers 3

        AudioQueueRef mQueue;
        AudioStreamBasicDescription mFileFormat;
        AudioQueueBufferRef mBuffers[ kNumberBuffers];
        UInt32 mBufferByteSize;
        Float32 mVolume;

        std::vector<AudioQueueBufferRef> mBuffersToDispose;

        FILE *mOggFile;
        OggVorbis_File mOggStream;
        vorbis_info *mVorbisInfo;
        vorbis_comment *mVorbisComment;
        ALenum mFormat;

    public:

        AudioQueueOggStream();
        ~AudioQueueOggStream();

        const char *errorString( int inCode);
        void open( const char *inPath);
        void display();

        void Teardown();
        static void QueueStoppedProc( void *inUserData, AudioQueueRef inAQ, AudioQueuePropertyID inID);
        static Boolean DisposeBuffer( AudioQueueRef inAQ, std::vector<AudioQueueBufferRef> inDisposeBufferList, AudioQueueBufferRef inBufferToDispose);
        static void QueueCallback( void *inUserData, AudioQueueRef inAQ, AudioQueueBufferRef inCompleteAQBuffer);
        void SetupQueue();
        void SetupBuffers();
        void Start();
        void Stop();
        void SetVolume( Float32 inVolume);
};

class AudioUnitOggStream
{   
    private:

        #define STREAM_BUFFER_SIZE              ( 1024 * 4)
        #define STREAM_BUFFER_COUNT             8
        #define STREAM_CIRCULAR_BUFFER_SIZE     ( STREAM_BUFFER_COUNT * STREAM_BUFFER_SIZE)

        enum OggStreamBufferState
        {
            kOggStreamBufferEmpty,
            kOggStreamBufferWriting,
            kOggStreamBufferFilled
        };

        typedef struct OggStreamBuffer
        {
            OggStreamBufferState mBufferState;
            int mBufferSize;
            int mReadPosition;
            char mBuffer[ STREAM_BUFFER_SIZE];
        } OggStreamBuffer;

        int mBuffersWritten;
        int mReadBufferIndex;
        int mWriteBufferIndex;
        OggStreamBuffer mBufferArray[ STREAM_BUFFER_COUNT];


        int mBufferBytesWritten;
        int mBufferReadPos;
        int mBufferWritePos;
        char mCircularBuffer[ STREAM_CIRCULAR_BUFFER_SIZE];

        static AudioUnitOggStream *sInstance;

        AudioComponentInstance mAudioUnit;
        AudioStreamBasicDescription mFileFormat;
        OSStatus mStatus;

        char mFileName[ 1024];
        bool mbLoadFile;
        bool mbCloseFile;
        bool mbAudioUnitRunning;
        bool mbThreadRunning;

        FILE *mOggFile;
        OggVorbis_File mOggStream;

        vorbis_info *mVorbisInfo;
        vorbis_comment *mVorbisComment;
        ALenum mFormat;

        StreamThread *mPlayThread;

    public:

        static void run();
        static void stopOgg();
        static void streamOgg( const char *inFileName);
        static OSStatus playbackCallback( void *inRefCon, AudioUnitRenderActionFlags *ioActionFlags, const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber, UInt32 inNumberFrames, AudioBufferList *ioData);
        static void *playbackProc( void *inData);

        AudioUnitOggStream();
        ~AudioUnitOggStream();

        void readDataFromBufferCircular( void *inRefCon, AudioUnitRenderActionFlags *ioActionFlags, const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber, UInt32 inNumberFrames, AudioBufferList *ioData);
        void readDataFromBuffer( void *inRefCon, AudioUnitRenderActionFlags *ioActionFlags, const AudioTimeStamp *inTimeStamp, UInt32 inBusNumber, UInt32 inNumberFrames, AudioBufferList *ioData);
        bool writeDataToBufferCircular();
        bool writeDataToBuffer();

        const char *errorString( int inCode);
        void display();
        void open( const char *inFileName);
        void close();

        void teardown();
        void checkStatus();
        void start();
        void stop();
};

#endif

